{-----------------------------------------------------------------------------
Picture conversion algorithm - ported from SCRPLUS (C) Edward Cree 2010
from C to Delphi.

------------------------------------------------------------------------------
       RENDERER.PAS - The renderer of Image2ULAPlus. Converts a picture
------------------------------------------------------------------------------

This file is part of Image2ULAPlus.

Image2ULAPlus is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Image2ULAPlus is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
------------------------------------------------------------------------------
The Initial Developer of the Original ULAPlus conversion algorithm is
Edward Cree (C) Copyright 2010. Portions created by Klaus Jahn are
Copyright (C) 2011. All Rights Reserved.

Contributor(s): Edward Cree.

You may retrieve the latest version of this file at the ZX-Modules home page,
located at http://www.zxmodules.de

-------------------------------------------------------------------------------
The following modifications have been made in comparison to the original
ScrPlus algorithm:

- Ability to show a progressbar during the rendering time

- Putting all that has to do with picture conversion into an independant
  object called ULAPlusRenderer.

- Converting the members of the structured C variable "ALL" into useful
  Delphi records for faster array access

- Removed the unused array YUVO.

- Changing the calculation factor of the RGB to YUV calculation according
  to the YUV calculation description found on
  http://www.fourcc.org/fccyvrgb.php

- Making the algorithm size-independant. Not limited to 256x192 pixels.
  This is useful for ZXP file support which allows pictures with other sizes.

- Timex mode is active when ATTRIBUTEHEIGHT=1 instead of 8.

Cross-referenced components and libraries that are not part of this source:

- Delphi 2006 components (c) 1995-2004 Borland Software Corporation
-----------------------------------------------------------------------------}

Unit renderer;

Interface

Uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs,
  ulaplus, yuv_helper;

Const
  tap_ulapluspalette_id=#19#0#0#0#20#1'64colour'#151#0#0#0#151#0#31#153#0+
    #255#0#0#147#0#249#192'(('#190#176'"23635"+'#176'"256"*'+
    #190#176'"23636")+'#176'"48"):'#239'"":'#234#243#33#38#0#9#1#59#191#62#64#237+
    #121#62#1#6#255#237#121#175#6#191#237#121#8#126#6#255#237#121#8#35#60#254#64+
    #32#239#251#201;

  tap_ulapluspalette_intro:String=tap_ulapluspalette_id;

Type
  Pixels_Standard=Array[0..2,0..7,0..7,0..31] Of Byte;
    //Standard SCREEN with 256x192 pixels (6144 bytes)

  Attribs_Standard=Array[0..23,0..31] Of Byte;
    //Standard attributes with 24x32 bytes (768 bytes)

  Attribs_Timex=Array[0..2,0..7,0..7,0..31] Of Byte;
    //Timex Attributes with 192x32 attributes (6144 bytes)

  TStandardScreen=Packed Record
    pixels:Pixels_Standard;
    attrs:Attribs_Standard;
  End; {Record}

  TULAPlusStandardScreen=Packed Record
    pixels:Pixels_Standard;
    attrs:Attribs_Standard;
    Palette:Tulapluscolors;
  End; {Record}

  TTimexScreen=Packed Record
    pixels:Pixels_Standard;
    attrs:Attribs_Timex;
  End; {Record}

  TULAPlusTimexScreen=Packed Record
    pixels:Pixels_Standard;
    attrs:Attribs_Timex;
    Palette:Tulapluscolors;
  End; {Record}

  TPaletteTape=Packed Record
    intro:Array[1..Length(tap_ulapluspalette_id)] Of Char;
    Palette:Tulapluscolors;
    suffix:Array[1..2] Of Char;
  End; {Record}

  TPictureProperties=Record
    ULAPlusMode:Boolean;
    TimexMode:Boolean;
  End; {record}

  TOutputType=(ot_none,ot_standardscreen,ot_timexscreen,ot_other);

  TULAPlusRenderer=Class
  Private
    FAttributeHeight:Integer;
    FMaxColours:Integer;
    FClutMask:Integer;
    FOriginalBitmap:TBitmap;
    Fyl,Fyh,Ful,Fuh,Fvl,Fvh:Integer;
    FStretch,FTimex,FCalculatePalette:Boolean;
    FDither,FDitherFlow:Boolean;
    FDepress:Extended;
    FOnUpdateBitmap:TNotifyEvent;
    FOnUpdatePalette:TNotifyEvent;
    FOnUpdateEdits:TNotifyEvent;
    FPaletteRGB:TULAPlusPalette;
    FInputFile,FOutputFile,FPaletteFile:String;
    FOutputPixels:TBitmap;
    FOutputAttrs:String;
    FOutputBitmap:TBitmap;
    FInputProperties:TPictureProperties;
    Procedure SetMaxColours(Value:Integer);
    Procedure SetClutMask(Value:Integer);
    Procedure SetYL(Value:Integer);
    Procedure SetYH(Value:Integer);
    Procedure SetUL(Value:Integer);
    Procedure SetUH(Value:Integer);
    Procedure SetVL(Value:Integer);
    Procedure SetVH(Value:Integer);
    Procedure SetStretch(Value:Boolean);
    Procedure SetTimex(Value:Boolean);
    Procedure SetCalculatePalette(Value:Boolean);
    Procedure SetDither(Value:Boolean);
    Procedure SetDitherFlow(Value:Boolean);
    Procedure SetDepress(Value:Extended);
    Function GetPaletteRGB(Index:Integer):TColor;
    Procedure Read_BMP_Image(FileName:String);
    Procedure Read_JPG_Image(FileName:String);
    Procedure Read_PNG_Image(FileName:String);
    Procedure Read_GIF_Image(FileName:String);
    Procedure Read_ZXP_Image(FileName:String);
    Procedure Read_SCR_Image(FileName:String);
    Function Write_BMP_Image(FileName:String):Boolean;
    Function Write_JPG_Image(FileName:String):Boolean;
    Function Write_PNG_Image(FileName:String):Boolean;
    Function Write_GIF_Image(FileName:String):Boolean;
    Function Write_ZXP_Image(FileName:String):Boolean;
    Function Write_SCR_Image(FileName:String):Boolean;
    Procedure ReadStandardScreen(fs:TFileStream);
    Procedure ReadTimexScreen(fs:TFileStream);
    Procedure ReadULAPlusStandardScreen(fs:TFileStream);
    Procedure ReadULAPlusTimexScreen(fs:TFileStream);
    Function WriteStandardScreen(fs:TFileStream):Boolean;
    Function WriteTimexScreen(fs:TFileStream):Boolean;
    Function WriteULAPlusStandardScreen(fs:TFileStream):Boolean;
    Function WriteULAPlusTimexScreen(fs:TFileStream):Boolean;
    Procedure Read_TAP_Palette(FileName: String);
    Procedure Read_RAW_Palette(FileName: String);
    Procedure DoUpdateBitmap;
    Procedure DoUpdatePalette;
    Procedure DoUpdateEdits;
    Procedure InputScreen(Var Pixels:Pixels_Standard; Var Attrs:Attribs_Standard; bmp:TBitmap); Overload;
    Procedure OutputScreen(Var Pixels:Pixels_Standard; Var Attrs:Attribs_Standard); Overload;
    Procedure InputScreen(Var Pixels:Pixels_Standard; Var Attrs:Attribs_Timex; bmp:TBitmap); Overload;
    Procedure OutputScreen(Var Pixels:Pixels_Standard; Var Attrs:Attribs_Timex); Overload;
    Function GetOutputType:TOutputType;
    Function GetOutputProperties:TPictureProperties;
  Public
    Property MaxColours:Integer Read FMaxColours Write SetMaxColours;
    Property ClutMask:Integer Read FClutMask Write SetClutMask;
    Property OnUpdateBitmap:TNotifyEvent Read FOnUpdateBitmap Write FOnUpdateBitmap;
    Property OnUpdatePalette:TNotifyEvent Read FOnUpdatePalette Write FOnUpdatePalette;
    Property OnUpdateEdits:TNotifyEvent Read FOnUpdateEdits Write FOnUpdateEdits;
    Property PaletteRGB[Index:Integer]:TColor Read GetPaletteRGB;
    Property YL:Integer Read FYL Write SetYL;
    Property YH:Integer Read FYH Write SetYH;
    Property UL:Integer Read FUL Write SetUL;
    Property UH:Integer Read FUH Write SetUH;
    Property VL:Integer Read FVL Write SetVL;
    Property VH:Integer Read FVH Write SetVH;
    Property Stretch:Boolean Read FStretch Write SetStretch;
    Property Timex:Boolean Read FTimex Write SetTimex;
    Property CalculatePalette:Boolean Read FCalculatePalette Write SetCalculatePalette;
    Property Depress:Extended Read FDepress Write SetDepress;
    Property Dither:Boolean Read FDither Write SetDither;
    Property DitherFlow:Boolean Read FDitherFlow Write SetDitherFlow;
    Property InputFile:String Read FInputFile Write FInputFile;
    Property OutputFile:String Read FOutputFile Write FOutputFile;
    Property PaletteFile:String Read FPaletteFile Write FPaletteFile;
    Property OutputType:TOutputType Read GetOutputType;
    Property OutputBitmap:TBitmap Read FOutputBitmap;
    Property InputProperties:TPictureProperties Read FInputProperties;
    Property OutputProperties:TPictureProperties Read GetOutputProperties;
    Constructor Create;
    Destructor Destroy; override;
    Function Render(creation:Boolean; AllowManualBreak:Boolean):Boolean;
    Procedure ReadImageFile(Const FileName:String);
    Function WriteImageFile(Const FileName:String):Boolean; Overload;
    Function WriteImageFile:Boolean; Overload;
    Procedure WriteImageToZXPFormat;
    Procedure ReadPaletteFile(Const FileName:String);
    Procedure GetAdjustedBitmap(DestBmp:TBitmap);
    Procedure GetPaletteBitmap(DestBmp:TBitmap);
    Procedure ResetValues;
    Procedure CheckCommandParams(Var gui:Boolean);
    Function GetInputFileFilters:String;
    Function GetOutputFileFilters:String;
    Function CheckDuplicateFile:Boolean;
  End; {Class}

Function SliderVal(position:Integer):Integer;
Function Tweak(Value:TYUVbyte; yl,yh,ul,uh,vl,vh:Integer):TYUVbyte; Overload;

Var
  ULAPlusRenderer:TULAPlusRenderer;

implementation

uses
  Math, Jpeg, GifImage, PngImage, zxpformat,
  progresses;

Const
  FileFilters:Array[1..6] Of Record
    description,filter:String;
  End=
    ((description:'ZX-Spectrum SCREEN$ (*.scr)';  filter:'*.scr'),
     (description:'Windows bitmaps (*.bmp)';      filter:'*.bmp'),
     (description:'JPEG images (*.jpg)';          filter:'*.jpg'),
     (description:'GIF images (*.gif)';           filter:'*.gif'),
     (description:'PNG images (*.png)';           filter:'*.png'),
     (description:'ZX-Paintbrush images (*.zxp)'; filter:'*.zxp'));

Type
  Tattribute=Packed Record
    att_ink,att_paper:Byte;
    optimal:Array[Boolean] Of TYUVintn;
  End; {Record}
    //encapsules all things that we want to do with the attributes

  Ppa_attribute=^Tpa_attribute;
    //Pointer needed to accelerate the accesses to attributes.

  Tpa_attribute=Packed Record
    c:TYUVint;  //former yuvc
    s:TYUVintn; //former yuvs
    h:Double;   //former yuvh
  End; {Record}
    //replaces three arrays YUVC, YUVS and YUVH which have the same ranges

Function SliderVal(position:Integer):Integer;
//Calculates the slider hyperbolic value
Begin
   Result:=128+Floor(32*SinH((position-127)/32));
End;

Function Tweak(Value,l,h:Integer):Integer; Overload;
Begin
  If h<>l Then Result:=IntToByte((Value-l)*255 div (h-l))
    Else Result:=Value;
End;

Function Tweak(Value:TYUVbyte; yl,yh,ul,uh,vl,vh:Integer):TYUVbyte;
//Calculates the tweak value of all sliders for Y, U and V value.
Begin
  Result.IY:=Tweak(Value.IY,yl,yh);
  Result.IU:=Tweak(Value.IU,ul,uh);
  Result.IV:=Tweak(Value.IV,vl,vh);
End;

Procedure TULAPlusRenderer.InputScreen(Var Pixels:Pixels_Standard;
  Var Attrs:Attribs_Standard; bmp:TBitmap);
//Drawing a standard Spectrum screen (6912 bytes) into destination bitmap:
Var
  Index,block,cx,x,cy,by,ay,attribute,pixelbyte:Integer;
  InkIndex,PaperIndex:Integer;
  p:PRGBQUAD;
Begin
  bmp.PixelFormat:=pf32Bit;
  bmp.SetSize(256,192);
  For block := 0 To 2 Do
    For by := 0 To 7 Do
      For cy := 0 To 7 Do Begin
        p:=bmp.ScanLine[block*64+cy*8+by];
        ay:=block*8+cy;
        For cx := 0 To 31 Do Begin
          pixelbyte:=Pixels[block,by,cy,cx];
          attribute:=Attrs[ay,cx];
          GetInkPaper(attribute,InkIndex,PaperIndex);
          For x := 0 To 7 Do Begin
            If pixelbyte And 128>0 Then Index:=InkIndex
              Else Index:=PaperIndex;
            p^:=colortoRGBQUAD(FPaletteRGB[Index]);
            Inc(p);
            pixelbyte:=pixelbyte Shl 1;
          End; {For}
        End; {For}
      End; {For}
End;

Procedure TULAPlusRenderer.InputScreen(Var Pixels:Pixels_Standard; Var Attrs:Attribs_Timex; bmp:TBitmap);
//Drawing a Timex screen (2x 6912 bytes) into destination bitmap:
Var
  Index,block,cx,x,cy,by,attribute,pixelbyte:Integer;
  InkIndex,PaperIndex:Integer;
  p:PRGBQUAD;
Begin
  bmp.PixelFormat:=pf32Bit;
  bmp.SetSize(256,192);
  For block := 0 To 2 Do
    For by := 0 To 7 Do
      For cy := 0 To 7 Do Begin
        p:=bmp.ScanLine[block*64+cy*8+by];
        For cx := 0 To 31 Do Begin
          pixelbyte:=Pixels[block,by,cy,cx];
          attribute:=Attrs[block,by,cy,cx];
          GetInkPaper(attribute,InkIndex,PaperIndex);
          For x := 0 To 7 Do Begin
            If pixelbyte And 128>0 Then Index:=InkIndex
              Else Index:=PaperIndex;
            p^:=colortoRGBQUAD(FPaletteRGB[Index]);
            Inc(p);
            pixelbyte:=pixelbyte Shl 1;
          End; {For}
        End; {For}
      End; {For}
End;

Function TULAPlusRenderer.GetOutputType;
//Checks what mode is the best for the output file. This depends on the
//current settings of the renderer.
Begin
  If assigned(FOutputPixels) And (Length(FOutputAttrs)>0) Then Begin
    If (FOutputPixels.Width=256) And (FOutputPixels.Height=192) Then Begin
      case Length(FOutputAttrs) Of
        32*24: Result:=ot_standardscreen;
        32*192:Result:=ot_timexscreen;
        Else Result:=ot_other;
      End; {case}
    End
      Else Result:=ot_other;
  End
    Else Result:=ot_none;
End;

Function TULAPlusRenderer.GetOutputProperties;
//Checks what kind of output picture format we have:
begin
  Result.ULAPlusMode:=Not IsStandardPalette(FPaletteRGB);
  Result.TimexMode:=Assigned(FOutputPixels) And
    (Length(FOutputAttrs)=(FOutputPixels.Width Div 8)*FOutputPixels.Height);
end;

Procedure TULAPlusRenderer.OutputScreen(Var Pixels:Pixels_Standard;
  Var Attrs:Attribs_Standard);
//Draws the rendered output bitmap into the Pixels and Attrs fields of a
//standard Spectrum screen.
Var
  aindex,block,cx,x,xx,y,cy,by,ay,pixelbyte:Integer;
Begin
  If OutputType=ot_standardscreen Then Begin
    For block := 0 To 2 Do
      For by := 0 To 7 Do
        For cy := 0 To 7 Do Begin
          y:=block*64+cy*8+by;
          ay:=block*8+cy;
          aindex:=ay*32+1;
          xx:=0;
          For cx := 0 To 31 Do Begin
            pixelbyte:=0;
            For x := 0 To 7 Do Begin
              pixelbyte:=pixelbyte Shl 1;
              If FOutputPixels.Canvas.Pixels[xx,y]=0 Then pixelbyte:=pixelbyte Or 1;
              Inc(xx);
            End; {For}
            Pixels[block,by,cy,cx]:=pixelbyte;
            If by=0 Then Attrs[ay,cx]:=Ord(FOutputAttrs[aindex+cx]);
          End; {For}
        End; {For}
  End; {If}
End;

Procedure TULAPlusRenderer.OutputScreen(Var Pixels:Pixels_Standard;
  Var Attrs:Attribs_Timex);
//Draws the rendered output bitmap into the Pixels and Attrs fields of a
//Timex screen.
Var
  aindex,block,cx,x,xx,y,cy,by,pixelbyte:Integer;
Begin
  If OutputType=ot_timexscreen Then Begin
    For block := 0 To 2 Do
      For by := 0 To 7 Do
        For cy := 0 To 7 Do Begin
          y:=block*64+cy*8+by;
          aindex:=y*32+1;
          xx:=0;
          For cx := 0 To 31 Do Begin
            pixelbyte:=0;
            For x := 0 To 7 Do Begin
              pixelbyte:=pixelbyte Shl 1;
              If FOutputPixels.Canvas.Pixels[xx,y]=0 Then pixelbyte:=pixelbyte Or 1;
              Inc(xx);
            End; {For}
            Pixels[block,by,cy,cx]:=pixelbyte;
            Attrs[block,by,cy,cx]:=Ord(FOutputAttrs[aindex+cx]);
          End; {For}
        End; {For}
  End; {If}
End;

{----------------- Object methods of TULAPlusRenderer ---------------------------}

Constructor TULAPlusRenderer.Create;
Begin
  inherited Create;
  ResetValues;
  FOriginalBitmap:=TBitmap.Create;
  FOutputBitmap:=TBitmap.Create;
End;

Destructor TULAPlusRenderer.Destroy;
Begin
  FreeAndNil(FOutputBitmap);
  FreeAndNil(FOutputPixels);
  FOutputAttrs:='';
  FreeAndNil(FOriginalBitmap);
  inherited Destroy;
End;

Procedure TULAPlusRenderer.ResetValues;
//Resets the values to default values (according to the defaults as done in
//ScrPlus.
Begin
  FMaxColours:=64;
  FClutMask:=15;
  FStretch:=True;
  FCalculatePalette:=True;
  FTimex:=False;
  FAttributeHeight:=8;
  FDepress:=0;
  FYL:=60;
  FYH:=194;
  FUL:=60;
  FUH:=194;
  FVL:=60;
  FVH:=194;
  SetStandardPalette(FPaletteRGB);
  FDither:=True;
  FDitherFlow:=False;
  DoUpdateEdits;
  DoUpdatePalette;
  DoUpdateBitmap;
End;

Procedure TULAPlusRenderer.ReadImageFile(Const FileName: String);
//Reads any image format and stores the content within the renderer.
Var
  Ext:String;
Begin
  If Length(FileName)>0 Then Begin
    If FileExists(FileName) Then Begin
      Ext:=ExtractFileExt(FileName);
      If CompareText(Ext,'.bmp')=0 Then Read_BMP_Image(FileName)
        Else
      If CompareText(Ext,'.jpg')=0 Then Read_JPG_Image(FileName)
        Else
      If CompareText(Ext,'.gif')=0 Then Read_GIF_Image(FileName)
        Else
      If CompareText(Ext,'.png')=0 Then Read_PNG_Image(FileName)
        Else
      If CompareText(Ext,'.zxp')=0 Then Read_ZXP_Image(FileName)
        Else
      If CompareText(Ext,'.scr')=0 Then Read_SCR_Image(FileName)
        Else showmessage('File type not supported!');
    End
      Else ShowMessage('Image file not found!');
  End; {If}
End;

Function TULAPlusRenderer.WriteImageFile(Const FileName: String):Boolean;
//Outputs the rendered image into an output file.
Var
  Ext:String;
Begin
  Result:=False;
  If Length(FileName)>0 Then Begin
    Ext:=ExtractFileExt(FileName);
    If CompareText(Ext,'.bmp')=0 Then Result:=write_BMP_Image(FileName)
      Else
    If CompareText(Ext,'.jpg')=0 Then Result:=write_JPG_Image(FileName)
      Else
    If CompareText(Ext,'.png')=0 Then Result:=write_PNG_Image(FileName)
      Else
    If CompareText(Ext,'.gif')=0 Then Result:=write_GIF_Image(FileName)
      Else
    If Comparetext(Ext,'.zxp')=0 Then Result:=Write_ZXP_Image(FileName)
      Else
    If CompareText(Ext,'.scr')=0 Then Result:=Write_SCR_Image(FileName)
      Else ShowMessage('File type not supported!');
  End
    Else ShowMessage('File not saved!');
End;

Function TULAPlusRenderer.WriteImageFile:Boolean;
//Outputs the rendered image with the internal filename (from command line)
begin
  Result:=False;
  If Length(FOutputFile)>0 Then Result:=WriteImageFile(FOutputFile)
    Else ShowMessage('Missing file name!');
end;

Function ReplaceExt(Const FileName:String; Const NewExt:String):String;
//Replaces any extension in FileName by another extension NewExt.
Var
  Ext:String;
Begin
  If Length(FileName)>0 Then Begin
    Ext:=ExtractFileExt(FileName);
    If Length(Ext)=0 Then Result:=FileName+NewExt
      Else Result:=Copy(FileName,1,Length(FileName)-Length(Ext))+NewExt;
  End
    Else Result:='';
End;

Procedure TULAPlusRenderer.WriteImageToZXPFormat;
begin
  If Length(FOutputFile)>0 Then Begin
    FOutputFile:=ReplaceExt(FOutputFile,'.zxp');
    WriteImageFile(FOutputFile);
  End
    Else ShowMessage('File not saved!');
end;

Procedure TULAPlusRenderer.ReadPaletteFile(Const FileName: String);
//Reads a TAP or RAW palette file.
Var
  Ext:String;
Begin
  If Length(FileName)>0 Then Begin
    If FileExists(FileName) Then Begin
      Ext:=ExtractFileExt(FileName);
      If CompareText(Ext,'.tap')=0 Then Read_TAP_Palette(FileName)
        Else Read_RAW_Palette(FileName);
    End
      Else ShowMessage('Palette file not found!');
  End; {If}
End;

Procedure TULAPlusRenderer.Read_BMP_Image(FileName: String);
//Reads a Windows bitmap into the renderer.
Var
  fs:TFileStream;
  bmp:TBitmap;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    fs.Position:=0;
    bmp:=TBitmap.Create;
    Try
      bmp.LoadFromStream(fs);
      FOriginalBitmap.Assign(bmp);
      DoUpdateBitmap;
      FInputFile:=FileName;
      FInputProperties.ULAPlusMode:=False;
      FInputProperties.TimexMode:=False;
    Finally
      bmp.Free;
    End; {Try}
  Finally
    fs.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Write_BMP_Image(FileName: String):Boolean;
//Writes the rendered picture as a Windows bitmap to disk.
Var
  bmp:TBitmap;
Begin
  bmp:=TBitmap.Create;
  Try
    bmp.Assign(FOutputBitmap);
    bmp.PixelFormat:=pf24bit;
    bmp.SaveToFile(FileName);
    Result:=True;
  Finally
    bmp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.Read_JPG_Image(FileName: String);
//Reads a JPEG file into the renderer.
Var
  fs:TFileStream;
  jpg:TJPEGImage;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    fs.Position;
    jpg:=TJPEGImage.Create;
    Try
      jpg.LoadFromStream(fs);
      FOriginalBitmap.Assign(jpg);
      DoUpdateBitmap;
      FInputFile:=FileName;
      FInputProperties.ULAPlusMode:=False;
      FInputProperties.TimexMode:=False;
    Finally
      jpg.free;
    End; {Try}
  Finally
    fs.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Write_JPG_Image(FileName: String):Boolean;
//Writes the rendered picture as a JPEG file to disk.
Var
  bmp:TBitmap;
  jpg:TJPEGImage;
Begin
  bmp:=TBitmap.Create;
  Try
    jpg:=TJPEGImage.Create;
    Try
      bmp.Assign(FOutputBitmap);
      bmp.PixelFormat:=pf24bit;
      jpg.Assign(bmp);
      jpg.SaveToFile(FileName);
      Result:=True;
    Finally
      jpg.Free;
    End; {Try}
  Finally
    bmp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.Read_PNG_Image(FileName: String);
//Reads a PNG file into the renderer.
Var
  fs:TFileStream;
  png:TPNGObject;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    fs.Position;
    png:=TPNGObject.Create;
    Try
      png.LoadFromStream(fs);
      FOriginalBitmap.Assign(png);
      DoUpdateBitmap;
      FInputFile:=FileName;
      FInputProperties.ULAPlusMode:=False;
      FInputProperties.TimexMode:=False;
    Finally
      png.free;
    End; {Try}
  Finally
    fs.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Write_PNG_Image(FileName: String):Boolean;
//Writes the rendered picture as a PNG file to disk.
Var
  bmp:TBitmap;
  png:TPNGObject;
Begin
  bmp:=TBitmap.Create;
  Try
    png:=TPNGObject.Create;
    Try
      bmp.Assign(FOutputBitmap);
      bmp.PixelFormat:=pf24bit;
      png.Assign(bmp);
      png.SaveToFile(FileName);
      Result:=True;
    Finally
      png.Free;
    End; {Try}
  Finally
    bmp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.Read_GIF_Image(FileName: String);
//Reads a GIF file into the renderer.
Var
  fs:TFileStream;
  gif:TGIFImage;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    fs.Position;
    gif:=TGIFImage.Create;
    Try
      gif.LoadFromStream(fs);
      FOriginalBitmap.Assign(gif);
      DoUpdateBitmap;
      FInputFile:=FileName;
      FInputProperties.ULAPlusMode:=False;
      FInputProperties.TimexMode:=False;
    Finally
      gif.free;
    End; {Try}
  Finally
    fs.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Write_GIF_Image(FileName: String):Boolean;
//Writes the rendered picture as a GIF file to disk.
Var
  bmp:TBitmap;
  gif:TGIFImage;
Begin
  bmp:=TBitmap.Create;
  Try
    gif:=TGIFImage.Create;
    Try
      bmp.Assign(FOutputBitmap);
      bmp.PixelFormat:=pf24bit;
      gif.Assign(bmp);
      gif.SaveToFile(FileName);
      Result:=True;
    Finally
      gif.Free;
    End; {Try}
  Finally
    bmp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.Read_ZXP_Image(FileName: String);
//Reads a ZXP (ZX-Paintbrush) image file into the renderer.
//This also reads an ULA Palette, if present.
Var
  sl:TStringList;
  bmp:TBitmap;
  transparent,timex:Boolean;
  newpalette:TULAPlusPalette;
  zxperror:zxp_errors;
Begin
  sl:=TStringList.Create;
  Try
    sl.LoadFromFile(FileName);
    bmp:=TBitmap.Create;
    Try
      zxperror:=readbitmapfromZXPformat(sl,bmp,transparent,True,timex,@newpalette);
      If zxperror=zxp_ok Then Begin
        FOriginalBitmap.Assign(bmp);
        FPaletteRGB:=newpalette;
        DoUpdateBitmap;
        DoUpdatePalette;
        FInputFile:=FileName;
        FInputProperties.ULAPlusMode:=IsStandardPalette(newpalette);
        FInputProperties.TimexMode:=timex;
      End; {If}
    Finally
      bmp.free;
    End; {Try}
  Finally
    sl.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Write_ZXP_Image(FileName: String):Boolean;
//Writes the rendered picture into a ZXP file to disk.
Var
  zxp:TStringList;
  zxperror:zxp_errors;
Begin
  Result:=False;
  zxp:=TStringList.Create;
  Try
    zxperror:=writebitmaptozxpformat(FOutputPixels,FOutputAttrs,zxp,True,FPaletteRGB);
    If zxperror=zxp_ok Then Begin
      zxp.SaveToFile(FileName);
      Result:=True;
    End; {if}
  Finally
    zxp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.ReadStandardScreen(fs:TFileStream);
//Reads a standard Spectrum screen (6912 bytes) from the FileStream and stores
//it as a bitmap inside the renderer.
Var
  Screen:TStandardScreen;
  bmp:TBitmap;
Begin
  bmp:=TBitmap.Create;
  Try
    If fs.Read(Screen,SizeOf(Screen))=SizeOf(Screen) Then Begin
      SetStandardPalette(FPaletteRGB);
      InputScreen(Screen.pixels,Screen.attrs,bmp);
      FOriginalBitmap.Assign(bmp);
      DoUpdateBitmap;
    End; {If}
  Finally
    bmp.Free;
  End; {Try}
End;

Function TULAPlusRenderer.WriteStandardScreen(fs:TFileStream):Boolean;
//Writes the rendered screen data into the FileStream using standard Spectrum
//format.
Var
  Screen:TStandardScreen;
Begin
  Result:=False;
  OutputScreen(Screen.pixels,Screen.attrs);
  If fs.Write(Screen,SizeOf(Screen))=SizeOf(Screen) Then Result:=True
    Else ShowMessage('File write error!');
End;

Procedure TULAPlusRenderer.ReadTimexScreen(fs:TFileStream);
//Reads a Timex screen (6912 bytes) from the FileStream and stores
//it as a bitmap inside the renderer.
Var
  Screen:TTimexScreen;
  bmp:TBitmap;
Begin
  bmp:=TBitmap.Create;
  Try
    If fs.Read(Screen,SizeOf(Screen))=SizeOf(Screen) Then Begin
      SetStandardPalette(FPaletteRGB);
      InputScreen(Screen.pixels,Screen.attrs,bmp);
      FOriginalBitmap.Assign(bmp);
      DoUpdateBitmap;
    End; {If}
  Finally
    bmp.Free;
  End; {Try}
End;

Function TULAPlusRenderer.WriteTimexScreen(fs:TFileStream):Boolean;
//Writes the rendered screen data into the FileStream using Timex
//format.
Var
  Screen:TTimexScreen;
Begin
  Result:=False;
  OutputScreen(Screen.pixels,Screen.attrs);
  If fs.Write(Screen,SizeOf(Screen))=SizeOf(Screen) Then Result:=True
    Else ShowMessage('File write error!');
End;

Procedure TULAPlusRenderer.ReadULAPlusStandardScreen(fs:TFileStream);
//Reads a standard Spectrum screen file containing a palette addition into
//the renderer.
Var
  Screen:TULAPlusStandardScreen;
  bmp:TBitmap;
Begin
  bmp:=TBitmap.Create;
  Try
    If fs.Read(Screen,Sizeof(Screen))=SizeOf(Screen) Then Begin
      SetPalette(FPaletteRGB,Screen.Palette);
      InputScreen(Screen.pixels,Screen.attrs,bmp);
      FOriginalBitmap.Assign(bmp);
      SetPalette(FPaletteRGB,Screen.Palette);
      DoUpdateBitmap;
      DoUpdatePalette;
    End; {If}
  Finally
    bmp.Free;
  End; {Try}
End;

Function TULAPlusRenderer.WriteULAPlusStandardScreen(fs:TFileStream):Boolean;
//Writes the rendered screen data plus current ULA Plus palette into the
//FileStream.
Var
  Screen:TULAPlusStandardScreen;
Begin
  Result:=False;
  OutputScreen(Screen.pixels,Screen.attrs);
  SetPalette(Screen.Palette,FPaletteRGB);
  If fs.Write(Screen,SizeOf(Screen))=SizeOf(Screen) Then Result:=True
    Else ShowMessage('File write error!');
End;

Procedure TULAPlusRenderer.ReadULAPlusTimexScreen(fs:TFileStream);
//Reads a Timex screen file containing a palette addition into the renderer.
Var
  Screen:TULAPlusTimexScreen;
  bmp:TBitmap;
Begin
  bmp:=TBitmap.Create;
  Try
    If fs.Read(Screen,SizeOf(Screen))=SizeOf(Screen) Then Begin
      SetPalette(FPaletteRGB,Screen.Palette);
      InputScreen(Screen.pixels,Screen.attrs,bmp);
      FOriginalBitmap.Assign(bmp);
      SetPalette(FPaletteRGB,Screen.Palette);
      DoUpdateBitmap;
      DoUpdatePalette;
    End; {If}
  Finally
    bmp.Free;
  End; {Try}
End;

Function TULAPlusRenderer.WriteULAPlusTimexScreen(fs:TFileStream):Boolean;
//Writes the rendered screen data plus current ULAPlus palette into the
//FileStream using Timex format.
Var
  Screen:TULAPlusTimexScreen;
Begin
  Result:=False;
  OutputScreen(Screen.pixels,Screen.attrs);
  SetPalette(Screen.Palette,FPaletteRGB);
  If fs.Write(Screen,SizeOf(Screen))=SizeOf(Screen) Then Result:=True
    Else ShowMessage('File write error!');
End;

Procedure TULAPlusRenderer.Read_SCR_Image(FileName: String);
//Reads any SCR file into the renderer.
Var
  fs:TFileStream;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    fs.Position:=0;
    //Depending on the file size, now load the corresponding Screen$ And
    //convert them To Bitmap format:
    case fs.Size Of
      SizeOf(TStandardScreen):Begin
        ReadStandardScreen(fs);
        FInputFile:=FileName;
        FInputProperties.ULAPlusMode:=False;
        FInputProperties.TimexMode:=False;
      End; {Standard SCREEN}
      SizeOf(TULAPlusStandardScreen):Begin
        ReadULAPlusStandardScreen(fs);
        FInputFile:=FileName;
        FInputProperties.ULAPlusMode:=True;
        FInputProperties.TimexMode:=False;
      End; {ULAPlus Standard SCREEN}
      SizeOf(TTimexScreen):Begin
        ReadTimexScreen(fs);
        FInputFile:=FileName;
        FInputProperties.ULAPlusMode:=False;
        FInputProperties.TimexMode:=True;
      End; {Timex Standard SCREEN}
      SizeOf(TULAPlusTimexScreen):Begin
        ReadULAPlusTimexScreen(fs);
        FInputFile:=FileName;
        FInputProperties.ULAPlusMode:=True;
        FInputProperties.TimexMode:=True;
      End; {Timex ULAPlus SCREEN}
      Else ShowMessage('Illegal size of SCR file!');
    End; {case}
  Finally
    fs.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Write_SCR_Image(FileName: String):Boolean;
//Writes the rendered screen into a SCR file and taking care about the
//specific screen type.
Var
  fs:TFileStream;
  ot:TOutputType;
Begin
  Result:=False;
  ot:=OutputType;
  If ot In [ot_standardscreen,ot_timexscreen] Then Begin
    fs:=TFileStream.Create(FileName,fmCreate);
    Try
      If IsStandardPalette(FPaletteRGB) Then Begin
        case ot Of
          ot_standardscreen:Result:=WriteStandardScreen(fs);
          ot_timexscreen:   Result:=WriteTimexScreen(fs);
        End; {case}
      End
        Else Begin
          case ot Of
            ot_standardscreen:Result:=WriteULAPlusStandardScreen(fs);
            ot_timexscreen:   Result:=WriteULAPlusTimexScreen(fs);
          End; {case}
        End; {Else}
    Finally
      fs.Free;
    End; {Try}
  End
    Else ShowMessage('File cannot be saved as SCR format!');
End;

Procedure TULAPlusRenderer.Read_TAP_Palette(FileName: String);
//Reads a TAP palette file and updates the palette display.
Var
  fs:TFileStream;
  palettecontent:TPaletteTape;
  i:Integer;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    fs.Position:=0;
    //File must have exactly 176 bytes:
    If fs.Size=SizeOf(palettecontent) Then Begin
      If fs.Read(palettecontent,SizeOf(palettecontent))=SizeOf(palettecontent) Then Begin
        If palettecontent.intro=tap_ulapluspalette_intro Then Begin
          For i := 0 To 63 Do
            FPaletteRGB[i]:=getRGBfromColor64(palettecontent.Palette[i]);
          DoUpdatePalette;
          FPaletteFile:=FileName;
        End
          Else ShowMessage('Invalid palette header!');
      End
        Else ShowMessage('Palette file read error!');
    End
      Else ShowMessage('Tape file must have exactly 176 bytes!');
  Finally
    fs.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.Read_RAW_Palette(FileName: String);
//Reads a RAW palette file and updates the palette display.
Var
  fs:TFileStream;
  Palette:TULApluscolors;
  i:Integer;
Begin
  fs:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyNone);
  Try
    //File must have exactly 64 bytes:
    If fs.Size=SizeOf(Palette) Then Begin
      If fs.Read(Palette,SizeOf(Palette))=SizeOf(Palette) Then Begin
        For i := 0 To 63 Do
          FPaletteRGB[i]:=getRGBfromColor64(Palette[i]);
        DoUpdatePalette;
        FPaletteFile:=FileName;
      End
        Else ShowMessage('Palette file read error!');
    End
      Else ShowMessage('Palette file must have exactly 64 bytes!');
  Finally
    fs.Free;
  End; {Try}
End;

Function TULAPlusRenderer.Render;
  //results in False if the user performs a manual break
Var
  pydist:^Integer;
  e:Extended;
  yuv:Array Of Array Of TYUVbyte;
  attributes:Array Of Tattribute;
  pa_attributes:Array Of Tpa_attribute;
  ydist:Array[0..255] Of Integer;
  Palette:Array[0..63] Of TYUVint;
  palm:Array[0..63] Of Ppa_attribute;
  pai,paj,pak,pav,pan:Ppa_attribute;
  Progress,color256,cwidth,cheight,cx,cxx,cy,cyy,x,y,i,j,k,v,Index,ColorCount:Integer;
  clut:Integer;
  inv:Boolean;
  pyuv,piyuv,ppyuv:TYUVbyte;
  lr:Array[Boolean] Of Record
    sx:Extended;
    sm,c:Integer;
  End; {Record}
  dinkYUV,dpaperYUV,oyuv,dyuv,qe:TYUVint;
  sn,ink,paper:Array[Boolean] Of Integer;
  dist:Array[Boolean] Of Double;
  syuv:Array[Boolean] Of TYUVreal;
  bx,by,PaletteEntries:Integer;
  smv,sink,spaper:Integer;
  m,sxm,sxs,paih,pakh,dd,di,dp,factor:Extended;
  p:^RGBQUAD;
  ManualBreak:Boolean;
  color:TColor;
  pstart:Integer;
  pixmap,sourcebmp:TBitmap;
  header:String;
Const
  pixelcolor:Array[Boolean] Of TColor=(clWhite,clBlack);
  pindex:Array[0..7] Of Integer=(8,0,24,16,40,32,56,48);
Begin
  Result:=True;
  sourcebmp:=TBitmap.Create;
  Try
    GetAdjustedBitmap(sourcebmp);
    sourcebmp.PixelFormat:=pf32bit;
    cwidth:=sourcebmp.Width div 8;
    cheight:=sourcebmp.Height div FAttributeHeight;
    //Ensure that there's no invalid width/height:
    sourcebmp.SetSize(8*cwidth,FAttributeHeight*cheight);
    Progress:=0;
    ManualBreak:=False;
    If FCalculatePalette Then header:='Creating ULAPlus picture...'
      Else header:='Creating ULAPlus picture and Palette...';
    ProgressBarForm.init(cheight*cwidth+2*cheight*cwidth,header,AllowManualBreak);
    Try
      //reading the source bitmap and calculating the YUV values Of each pixel
      //into the Array YUV - Step 2 in the original C algorithm.
      //Resampling is not necessary, because the algorithm is
      //working for any size Of the bitmap with (width mod 8) =0 and
      //(height mod FAttributeHeight) =0 :

      SetLength(yuv,sourcebmp.Height,sourcebmp.Width);
      Try
        pixmap:=TBitmap.Create;
        Try
          pixmap.PixelFormat:=pf1bit;
          pixmap.SetSize(sourcebmp.Width,sourcebmp.Height);
          For y := 0 To sourcebmp.height-1 Do Begin
            p:=sourcebmp.ScanLine[y];
            For x := 0 To sourcebmp.Width - 1 Do Begin
              yuv[y,x]:=getYUVfromRGB(p^);
              Inc(p);
            End; {For}
          End; {For}

          //Per-block YUV block conversion (Step 3 in the original C algorithm):
          SetLength(attributes,cheight*cwidth);
          Try
            Index:=0;
            cy:=0;
            For by := 0 To cheight-1 Do Begin
              cx:=0;
              For bx := 0 To cwidth-1 Do Begin
                Inc(Progress);
                With attributes[Index] Do Begin
                  FillChar(ydist,sizeof(ydist),0);
                  cyy:=cy;
                  ProgressBarForm.Progress(Progress);
                  If AllowManualBreak Then Begin
                    ManualBreak:=ProgressBarForm.ManualBreak;
                    If ManualBreak Then Begin
                      Result:=False;
                      Exit;
                    End; {If}
                  End; {If}
                  For y := 0 To FAttributeHeight - 1 Do Begin
                    cxx:=cx;
                    For x := 0 To 7 Do Begin
                      i:=yuv[cyy,cxx].IY;
                      Inc(ydist[i]);
                      Inc(cxx);
                    End; {For}
                    Inc(cyy);
                  End; {For}
                  sxm:=-1;
                  smv:=yuv[cy,cx].IY;
                  For i := 0 To 255 Do Begin
                    FillChar(lr,sizeof(lr),0);
                    pydist:=@ydist[0];
                    For j := 0 To 255 Do Begin
                      With lr[j>i] Do Begin
                        sx:=sx+pydist^*j;
                        Inc(c,pydist^);
                      End; {With}
                      Inc(pydist);
                    End; {For}
                    For inv := False To True Do Begin
                      With lr[inv] Do Begin
                        If c<>0 Then sm:=Round(sx/c)
                          Else sm:=0;
                        sx:=0;
                      End; {With}
                    End; {For}
                    pydist:=@ydist[0];
                    For j := 0 To 255 Do Begin
                      With lr[j>i] Do
                        sx:=sx+pydist^*(j-sm)*(j-sm);
                      Inc(pydist);
                    End; {For}
                    sxs:=lr[False].sx+lr[True].sx;
                    If (sxs<sxm) Or (sxm<0) Then Begin
                      sxm:=sxs;
                      smv:=i;
                    End; {If}
                  End; {For}
                  // smv is now our 'splitpoint' between ink And paper
                  // calculate optimal y-u-v
                  FillChar(syuv,sizeof(syuv),0);
                  FillChar(sn,sizeof(sn),0);
                  cyy:=cy;
                  For y := 0 To FAttributeHeight-1 Do Begin
                    cxx:=cx;
                    For x := 0 To 7 Do Begin
                      pyuv:=yuv[cyy,cxx];
                      inv:=pyuv.IY>smv;
                      addYUV(syuv[inv],pyuv);
                      Inc(sn[inv]);
                      Inc(cxx);
                    End; {For}
                    Inc(cyy);
                  End; {For}
                  For inv := False To True Do
                    With syuv[inv],optimal[inv] Do Begin
                      i:=max(1,sn[inv]);
                      IY:=Round(RY/i);
                      IU:=Round(RU/i);
                      IV:=Round(RV/i);
                      nn:=sn[inv];
                    End; {With}
                End; {With}
                Inc(Index);
                Inc(cx,8);
              End; {For}
              Inc(cy,FAttributeHeight);
            End; {For}

            //Palette creation (Step 4 in the original C algorithm):
            If FCalculatePalette Then Begin
              SetLength(pa_attributes,cwidth*cheight*2);
              Try
                FillChar(Palette,sizeof(Palette),0);
                Index:=0;
                pai:=@pa_attributes[0];
                For by := 0 To cheight-1 Do Begin
                  For bx := 0 To cwidth-1 Do Begin
                    With attributes[Index] Do Begin
                      For inv := False To True Do Begin
                        pai.s:=optimal[inv];
                        Inc(pai);
                      End; {For}
                    End; {With}
                    Inc(Index);
                  End; {For}
                End; {For}
                ColorCount:=0;
                m:=0;
                pav:=@pa_attributes[0];
                pai:=@pa_attributes[0];
                paj:=@pa_attributes[0];
                For i := 0 To 2*cwidth*cheight-1 Do Begin
                  Inc(Progress);
                  ProgressBarForm.Progress(Progress);
                  If AllowManualBreak Then Begin
                    ManualBreak:=ProgressBarForm.ManualBreak;
                    If ManualBreak Then Begin
                      Result:=False;
                      Exit;
                    End; {If}
                  End; {If}
                  If pai.s.nn>0 Then Begin
                    paj.c:=ConvertYUV(pai.s);
                    paj.h:=pai.s.nn;
                    pak:=pai;
                    Inc(pak);
                    For k := i+1 To 2*cwidth*cheight-1 Do Begin
                      If compareYUV(pai.s,pak.s) Then Begin
                        paj.h:=paj.h+pak.s.nn;
                        pak.s.nn:=0;
                      End; {If}
                      Inc(pak);
                    End; {For}
                    If paj.h>m Then Begin
                      m:=paj.h;
                      pav:=paj;
                    End; {If}
                    Inc(ColorCount);
                    Inc(paj);
                  End; {If}
                  Inc(pai);
                End; {For}

                // ColorCount is now the # Of relevant color entries, now
                // pump up the wantedness Of closely clustered colors
                ProgressBarForm.Progress(0);
                ProgressBarForm.SetMax(ColorCount+2*cwidth*cheight);
                Progress:=0;
                pai:=@pa_attributes[0];
                For i := 0 To ColorCount - 1 Do Begin
                  Inc(Progress);
                  ProgressBarForm.Progress(Progress);
                  If AllowManualBreak Then Begin
                    ManualBreak:=ProgressBarForm.ManualBreak;
                    If ManualBreak Then Begin
                      Result:=False;
                      Exit;
                    End; {If}
                  End; {If}
                  pak:=@pa_attributes[0];
                  For k := 0 To ColorCount - 1 Do Begin
                    If i<>k Then Begin
                      dd:=Sqrt(squareYUV(deltaYUV(pai.c,pak.c)));
                      If dd<>0 Then Begin
                        paih:=pai.h;
                        pakh:=pak.h;
                        factor:=0.002/dd;
                        pai.h:=pai.h+pakh*factor;
                        pak.h:=pak.h+paih*factor;
                      End; {If}
                    End; {If}
                    Inc(pak);
                  End; {For}
                  Inc(pai);
                End; {For}

                //Generating the palette entries:
                PaletteEntries:=0;
                while PaletteEntries<FMaxColours Do Begin
                  clut:=PaletteEntries Shr 4;
                  If (1 Shl clut) And Fclutmask<>0 Then Begin
                    palm[PaletteEntries]:=pav;
                    Palette[PaletteEntries]:=pav.c;
                    Inc(PaletteEntries);
                    pav.h:=0;
                    pan:=@pa_attributes[0];
                    m:=0;
                    pai:=@pa_attributes[0];
                    For i := 0 To ColorCount - 1 Do Begin
                      dyuv:=deltaYUV(pai.c,pav.c);
                      With dyuv Do Begin
                        e:=(IY/(2+IY))+(IU/(2+IU))+(IV/(2+IV));
                        If FDepress<>0 Then pai.h:=pai.h*Power(e,Exp(FDepress))
                          Else pai.h:=pai.h*e;
                      End; {With}
                      If pai.h>m Then Begin
                        m:=pai.h;
                        pan:=pai;
                      End; {If}
                      Inc(pai);
                    End; {For}
                    pav:=pan;
                  End
                    Else Inc(PaletteEntries);
                End; {while}
              Finally
                SetLength(pa_attributes,0);
              End; {Try}

              //writing the palette into the palette bytes:
              For i := 0 To 63 Do
                FpaletteRGB[i]:=getcolorfromYUV(Palette[i]);
               DoUpdatePalette;
            End
              Else Begin
                //No palette has to be created - RGBPalette is left unchanged.
                //Instead we have to read the palette into the internal palette
                For i := 0 To 63 Do
                  Palette[i]:=getYUVfromColor(FPaletteRGB[i]);
              End; {Else}
            //Dithering the 8x8 or 8x1 pixel map using Floyd-Steinberg algorithm:
            //which is weight-calculated like this (Step 5 Of the original C code):
            // 1/16 * | 0  X  7 |
            //        | 3  5  1 |
            //where X is the current pixel position.
            Index:=0;
            cy:=0;
            For by := 0 To cheight-1 Do Begin
              cx:=0;
              For bx := 0 To cwidth - 1 Do Begin
                With attributes[Index] Do Begin
                  For inv := False To True Do Begin
                    Inc(Progress);
                    ProgressBarForm.Progress(Progress);
                    If AllowManualBreak Then Begin
                      ManualBreak:=ProgressBarForm.ManualBreak;
                      If ManualBreak Then Begin
                        Result:=False;
                        Exit;
                      End; {If}
                    End; {If}
                    oyuv:=ConvertYUV(optimal[inv]);
                    m:=-1;
                    v:=0;
                    For i := 0 To FMaxColours-1 Do Begin
                      clut:=i Shr 4;
                      If (1 Shl clut) And Fclutmask<>0 Then Begin
                        dd:=Sqrt(squareYUV(deltaYUV(oyuv,Palette[i])));
                        If (dd<m) Or (m<0) Then Begin
                          m:=dd;
                          v:=i;
                        End; {If}
                      End; {If}
                    End; {For}
                    dist[inv]:=m*Sqrt(optimal[inv].nn);
                    ink[inv]:=v;
                    oyuv:=ConvertYUV(optimal[Not inv]);
                    M:=-1;
                    v:=0;
                    pstart:=pindex[ink[inv] div 8];
                    For i := pstart To pstart+7 Do Begin
                      dd:=Sqrt(squareYUV(deltaYUV(oyuv,Palette[i])));
                      If (dd<m) Or (m<0) Then Begin
                        m:=dd;
                        v:=i;
                      End; {If}
                    End; {For}
                    dist[inv]:=dist[inv]+m*Sqrt(optimal[Not inv].nn);
                    paper[inv]:=v;
                  End; {For}
                  If dist[True]<dist[False] Then Begin
                    sink:=ink[True];
                    spaper:=paper[True];
                  End
                    Else Begin
                      sink:=ink[False];
                      spaper:=paper[False];
                    End; {Else}
                  If sink mod 16<8 Then Begin
                    att_ink:=sink;
                    att_paper:=spaper;
                  End
                    Else Begin
                      att_paper:=sink;
                      att_ink:=spaper;
                    End; {Else}
                  cxx:=cx;
                  For x := 0 To 7 Do Begin
                    cyy:=cy;
                    For y := 0 To FAttributeHeight-1 Do Begin
                      pyuv:=yuv[cyy,cxx];
                      dinkyuv:=diffYUV(pyuv,Palette[att_ink]);
                      di:=squareYUV(dinkyuv);
                      dpaperyuv:=diffYUV(pyuv,Palette[att_paper]);
                      dp:=squareYUV(dpaperyuv);
                      pixmap.Canvas.Pixels[cxx,cyy]:=pixelcolor[di<dp];
                      If FDither Then Begin
                        If di<dp Then qe:=dinkyuv
                          Else qe:=dpaperyuv;
                        If x<7 Then Begin
                          addmulYUV(yuv[cyy,cxx+1],7/16,qe);
                          If y<FAttributeHeight-1 Then addmulYUV(yuv[cyy+1,cxx+1],1/16,qe);
                        End; {If}
                        If y<FAttributeHeight-1 Then Begin
                          addmulYUV(yuv[cyy+1,cxx],5/16,qe);
                          If cxx>0 Then addmulYUV(yuv[cyy+1,cxx-1],3/16,qe);
                        End; {If}
                        If FDitherFlow Then Begin
                          If (x=7) And (bx<cwidth-1) And ((y mod 2)<>0) Then Begin
                            With yuv[cyy,cxx+1] Do Begin
                              IY:=pyuv.IY-qe.IY;
                              IU:=pyuv.IU-qe.IU;
                              IV:=pyuv.IV-qe.IV;
                            End; {With}
                          End; {If}
                        End; {If}
                      End; {If}
                      Inc(cyy);
                    End; {For}
                    Inc(cxx);
                  End; {For}
                End; {With}
                Inc(Index);
                Inc(cx,8);
              End; {For}
              Inc(cy,FAttributeHeight);
            End; {For}

            //Writing the calculated pixels and attributes into the destination
            //bitmap and attribute String:
            If creation Then Begin
              If Not Assigned(FOutputPixels) Then FOutputPixels:=TBitmap.Create;
              FOutputPixels.Assign(pixmap);
              SetLength(FOutputAttrs,cheight*cwidth);
            End; {If}

            FOutputBitmap.Assign(sourcebmp);
            Index:=0;
            cy:=0;
            For by:=0 To cheight-1 Do Begin
              cx:=0;
              For bx := 0 To cwidth - 1 Do Begin
                With attributes[Index] Do Begin
                  cyy:=cy;
                  If creation Then Begin
                    FOutputAttrs[Index+1]:=Chr((att_ink mod 8)+8*(att_paper mod 8)+64*(att_ink div 16));
                  End; {If}
                  For y := 0 To FAttributeHeight-1 Do Begin
                    cxx:=cx;
                    For x := 0 To 7 Do Begin
                      If pixmap.Canvas.Pixels[cxx,cyy]=clBlack Then
                        color:=getcolorfromYUV(Palette[att_ink])
                        Else color:=getcolorfromYUV(Palette[att_paper]);
                      FOutputBitmap.Canvas.Pixels[cxx,cyy]:=color;
                      Inc(cxx);
                      Inc(p);
                    End; {For}
                    Inc(cyy);
                  End; {For}
                End; {With}
                Inc(Index);
                Inc(cx,8);
              End; {For}
              Inc(cy,FAttributeHeight);
            End; {For}
            DoUpdateBitmap;
            DoUpdateEdits;
          Finally
            SetLength(attributes,0);
          End;
        Finally
          pixmap.Free;
        End; {Try}
      Finally
        SetLength(yuv,0,0);
      End; {Try}
    Finally
      ProgressBarForm.done;
    End; {Try}
  Finally
    sourcebmp.Free;
  End; {try}
End;

Procedure TULAPlusRenderer.SetMaxColours(Value: Integer);
Begin
  If Value<>FMaxColours Then Begin
    If Value<8 Then Value:=8
      Else
    If Value>64 Then Value:=64;
    FMaxColours:=Value;
    DoUpdatePalette;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetClutMask(Value: Integer);
Begin
  If Value<>FClutMask Then Begin
    Value:=Value And 15;
    If Value<>0 Then Begin
      FClutMask:=Value;
      DoUpdatePalette;
    End; {If}
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetYL(Value:Integer);
Begin
  If Value<>FYL Then Begin
    FYL:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetYH(Value:Integer);
Begin
  If Value<>FYH Then Begin
    FYH:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetUL(Value:Integer);
Begin
  If Value<>FUL Then Begin
    FUL:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetUH(Value:Integer);
Begin
  If Value<>FUH Then Begin
    FUH:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetVL(Value:Integer);
Begin
  If Value<>FVL Then Begin
    FVL:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetVH(Value:Integer);
Begin
  If Value<>FVH Then Begin
    FVH:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetStretch(Value: Boolean);
Begin
  If Value<>FStretch Then Begin
    FStretch:=Value;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetTimex(Value: Boolean);
Begin
  If Value<>FTimex Then Begin
    FTimex:=Value;
    If Value Then Begin
      FStretch:=True;
      FAttributeHeight:=1;
    End
      Else FAttributeHeight:=8;
    DoUpdateBitmap;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetCalculatePalette(Value: Boolean);
Begin
  If Value<>FCalculatePalette Then Begin
    FCalculatePalette:=Value;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetDepress(Value: Extended);
Begin
  If Value<>FDepress Then Begin
    If Value<-1 Then Value:=-1
      Else
    If Value>1 Then Value:=1;
    If Value<>FDepress Then Begin
      FDepress:=Value;
      DoUpdateEdits;
    End; {If}
  End; {If}
End;

Procedure TULAPlusRenderer.SetDither(Value: Boolean);
Begin
  If Value<>FDither Then Begin
    FDither:=Value;
    DoUpdateEdits;
  End; {If}
End;

Procedure TULAPlusRenderer.SetDitherFlow(Value: Boolean);
Begin
  If Value<>FDitherFlow Then Begin
    FDitherFlow:=Value;
    DoUpdateEdits;
  End; {If}
End;

Function TULAPlusRenderer.GetPaletteRGB(Index:Integer):TColor;
Begin
  If (Index>=0) And (Index<=63) Then Result:=FPaletteRGB[Index]
    Else Result:=0;
End;

Procedure TULAPlusRenderer.GetAdjustedBitmap;
//Retrieves the original bitmap after YUV ruler adjustment.
Var
  y,x:Integer;
  p:PRGBQUAD;
  bmp:TBitmap;
  yuv:TYUVbyte;
  _yl,_yh,_ul,_uh,_vl,_vh:Integer;
Begin
  bmp:=TBitmap.Create;
  Try
    If FStretch And (FOriginalBitmap.Width>0) And (FOriginalBitmap.Height>0) Then Begin
      bmp.SetSize(256,192);
      setstretchbltmode(bmp.Canvas.Handle,HALFTONE);
      StretchBlt(bmp.Canvas.Handle,0,0,256,192,FOriginalBitmap.Canvas.Handle,0,0,
        FOriginalBitmap.Width,FOriginalBitmap.Height,SRCCOPY);
    End
      Else bmp.Assign(FOriginalBitmap);
    bmp.PixelFormat:=pf32bit;
    _yl:=SliderVal(yl);
    _yh:=SliderVal(yh);
    _ul:=SliderVal(ul);
    _uh:=SliderVal(uh);
    _vl:=SliderVal(vl);
    _vh:=SliderVal(vh);
    For y := 0 To bmp.Height-1 Do Begin
      p:=bmp.ScanLine[y];
      For x := 0 To bmp.Width-1 Do Begin
        yuv:=Tweak(getYUVfromRGB(p^),_yl,_yh,_ul,_uh,_vl,_vh);
        p^:=getRGBfromYUV(yuv);
        Inc(p);
      End; {For}
    End; {For}
    DestBmp.Assign(bmp);
  Finally
    bmp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.GetPaletteBitmap;
//Draws the palette into a bitmap.
Var
  bmp:TBitmap;
  x,y,xx,yy,Index,clut:Integer;
Begin
  bmp:=TBitmap.Create;
  Try
    bmp.Transparent:=False;
    bmp.SetSize(65,65);
    With bmp.Canvas Do Begin
      Brush.Style:=bsSolid;
      Brush.Color:=clBlack;
      FillRect(rect(0,0,65,65));
      Index:=0;
      yy:=1;
      For y := 0 To 7 Do Begin
        clut:=y Shr 1;
        xx:=1;
        For x := 0 To 7 Do Begin
          If ((1 Shl clut) And FClutMask>0) And (Index<FMaxColours) Then Begin
            Brush.Color:=FPaletteRGB[Index];
            FillRect(rect(xx,yy,xx+7,yy+7));
          End
            Else Begin
              Brush.Color:=clWhite;
              FillRect(rect(xx,yy,xx+7,yy+7));
              Pen.Color:=clBlack;
              MoveTo(xx,yy);
              LineTo(xx+7,yy+7);
              MoveTo(xx+7,yy);
              LineTo(xx,yy+7);
            End; {Else}
          Inc(Index);
          Inc(xx,8);
        End; {For}
        Inc(yy,8);
      End; {For}
    End; {With}
    DestBmp.Assign(bmp);
  Finally
    bmp.Free;
  End; {Try}
End;

Procedure TULAPlusRenderer.DoUpdateBitmap;
Begin
  If Assigned(FOnUpdateBitmap) Then FOnUpdateBitmap(Self);
End;

Procedure TULAPlusRenderer.DoUpdatePalette;
Begin
  If Assigned(FOnUpdatePalette) Then FOnUpdatePalette(Self);
End;

Procedure TULAPlusRenderer.DoUpdateEdits;
Begin
  If Assigned(FOnUpdateEdits) Then FOnUpdateEdits(Self);
End;

Procedure UnQuoteFileName(Var FileName:String);
//Results the filename without quotes "" that are used within command line
//to hold file names that may contain spaces.
Begin
  FileName:=Trim(FileName);
  If (Length(FileName)>2) And (FileName[1]='"') And (FileName[Length(FileName)]='"') Then
    FileName:=Trim(Copy(FileName,2,Length(FileName)-2))
End;

Function ExtractYUVValues(Const arg,Name:String; Var lo,hi:Integer):Boolean;
//Extracts the low and high value of one of the YUV-rulers.
Var
  p:Integer;
  s:String;
Begin
  Result:=False;
  If Length(arg)>0 Then Begin
    p:=Pos(',',arg);
    If p>0 Then Begin
      s:=Trim(Copy(arg,1,p-1));
      If Length(s)>0 Then Begin
        If TryStrToInt(s,lo) And (lo>=0) And (lo<=255) Then Begin
          s:=Copy(arg,p,Length(arg)-p+1);
          Delete(s,1,1);
          s:=Trim(s);
          If Length(s)>0 Then Begin
            If TryStrToInt(s,hi) And (hi>=0) And (hi<=255) Then Result:=True
              Else ShowMessage(Format('Invalid %s value for %s parameter (must be 0 .. 255)!',['high',Name]));
          End
            Else ShowMessage(Format('Missing %s value for %s parameter!',['high',Name]))
        End
          Else ShowMessage(Format('Invalid %s value for %s parameter (must be 0 .. 255)!',['low',Name]));
      End
        Else ShowMessage(Format('Missing %s value for %s parameter!',['low',Name]));
    End
      Else ShowMessage('Missing comma separator!');
  End
    Else ShowMessage(Format('Missing %s value for %s parameter!',['low',Name]));
End;

Procedure TULAPlusRenderer.CheckCommandParams;
//Analyses the command line and sets the values of the Renderer.
Type
  OutputModes=(om_none,om_screen,om_bitmap);
Var
  om:OutputModes;
  i,clut,lo,hi,err:Integer;
  e:Extended;
  arg:String;
Begin
  If ParamCount>0 Then Begin
    //Set SCRPlus standard parameters For CommandLine:
    Dither:=True;
    gui:=False;
    om:=om_none;
    For i := 1 To ParamCount Do Begin
      arg:=ParamStr(i);
      //Standard SCRPlus command parameters:
      If SameText('gui',arg) Then gui:=True
        Else
      If SameText('+t',arg) Then Timex:=True
        Else
      If SameText('-t',arg) Then Timex:=False
        Else
      If SameText('+d',arg) Then DitherFlow:=True
        Else
      If SameText('-d',arg) Then DitherFlow:=False
        Else
      If SameText('-s',arg) Then om:=om_screen
        Else
      If SameText('-b',arg) Then om:=om_bitmap
        Else
      If SameText('-i=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        UnQuoteFileName(arg);
        InputFile:=arg;
      End
        Else
      If SameText('-o=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        UnQuoteFileName(arg);
        OutputFile:=arg;
      End
        Else
      If SameText('-p=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        if SameText(arg,'none') then CalculatePalette:=False
          Else Begin
            UnQuoteFileName(arg);
            PaletteFile:=arg;
            CalculatePalette:=False;
          End; {Else}
      End
        Else
      If SameText('-c=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        arg:=Trim(arg);
        case Length(arg) Of
          0:ShowMessage('Missing clut mask!');
          1:Begin
            case arg[1] Of
              '1'..'9':ClutMask:=Ord(arg[1])-Ord('0');
              'a'..'f':ClutMask:=Ord(arg[1])-Ord('a')+10;
              'A'..'F':ClutMask:=Ord(arg[1])-Ord('A')+10;
              Else ShowMessage('Invalid clut mask (must be single hex digit 0..F or decimal number 0..15)!');
            End; {case}
          End; {1 digit}
          Else Begin
            If TryStrToInt(arg,clut) And (clut>=1) And (clut<=15) Then ClutMask:=clut
              Else ShowMessage('Invalid clut mask (must be single hex digit 0..F or decimal number 0..15)!');
          End; {2 digits}
        End; {case}
      End
        Else
      If SameText('-y=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        If ExtractYUVValues(Trim(arg),'Y',lo,hi) Then Begin
          YL:=lo;
          YH:=hi;
        End; {If}
      End
        Else
      If SameText('-u=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        If ExtractYUVValues(Trim(arg),'U',lo,hi) Then Begin
          UL:=lo;
          UH:=hi;
        End; {If}
      End
        Else
      If SameText('-v=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        If ExtractYUVValues(Trim(arg),'V',lo,hi) Then Begin
          VL:=lo;
          VH:=hi;
        End; {If}
      End
        Else
      If SameText('-e=',Copy(arg,1,3)) Then Begin
        Delete(arg,1,3);
        Val(Trim(arg),e,err);
        If (err=0) And (e>=-1) And (e<=1) Then Depress:=e
          Else ShowMessage('Invalid depress value (must be -1.0 .. 1.0)!');
      End
        Else
      //Special Image2ULAPlus command line parameters:
      If SameText('+scrsize',arg) Then Stretch:=True
        Else
      If SameText('-scrsize',arg) Then Stretch:=False;

    End; {For}
    //When no file has been specified, the GUI has to be opened:
    If (Length(InputFile)=0) Or (Length(OutputFile)=0) Then gui:=True;

    //Calculating Image Input file:
    If Length(InputFile)>0 Then ReadImageFile(InputFile);

    //Calculate image output file:
    If Length(OutputFile)>0 Then Begin
      case om Of
        om_screen:OutputFile:=ReplaceExt(OutputFile,'.scr');
        om_bitmap:OutputFile:=ReplaceExt(OutputFile,'.bmp');
      End; {case}
    End; {If}

    //Calculate palette file:
    If Length(PaletteFile)>0 Then ReadPaletteFile(PaletteFile);
  End
    Else gui:=True;
End;

Function TULAPlusRenderer.GetInputFileFilters:String;
Var
  i:Integer;
  allfilters:String;
Begin
  allfilters:='';
  For i:=Low(FileFilters) To High(FileFilters) Do Begin
    If Length(allfilters)>0 Then allfilters:=allfilters+';';
    allfilters:=allfilters+FileFilters[i].filter;
  End; {for}
  Result:='All compatible image files ('+allfilters+')|'+allfilters;
  For i:=Low(FileFilters) To High(FileFilters) Do Begin
    If Length(Result)>0 Then Result:=Result+'|';
    With FileFilters[i] Do
      Result:=Result+description+'|'+filter;
  End; {for}
End;

Function TULAPlusRenderer.GetOutputFileFilters:String;
Var
  i,start:Integer;
Begin
  Result:='';
  If GetOutputType In [ot_standardscreen,ot_timexscreen] Then start:=Low(FileFilters)
    Else start:=Succ(Low(FileFilters));
  For i:=start To High(FileFilters) Do Begin
    If Length(Result)>0 Then Result:=Result+'|';
    With FileFilters[i] Do
      Result:=Result+description+'|'+filter;
  End; {for}
End;

Function TULAPlusRenderer.CheckDuplicateFile;
begin
  If (Length(FOutputFile)>0) And FileExists(FOutputFile) Then Begin
    Result:=MessageDlg('Output file already exists. Overwrite (y/n)?',mtWarning,
      [mbYes,mbNo,mbCancel],0)=mrYes;
  End
    Else Result:=True;
end;

Initialization

  ULAPlusRenderer:=TULAPlusRenderer.Create;

Finalization

  FreeAndNil(ULAPlusRenderer);

End.
