{-----------------------------------------------------------------------------
Picture conversion algorithm - ported from SCRPLUS (C) Edward Cree 2010
from C to Delphi.

------------------------------------------------------------------------------
                ZXPFORMAT.PAS - Library for ZXP file support 
------------------------------------------------------------------------------

This file is part of Image2ULAPlus.

Image2ULAPlus is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Image2ULAPlus is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
------------------------------------------------------------------------------
The Initial Developer of the Original ULAPlus conversion algorithm is
Edward Cree (C) Copyright 2010. Portions created by Klaus Jahn are
Copyright (C) 2011. All Rights Reserved.

Contributor(s): Edward Cree.

You may retrieve the latest version of this file at the ZX-Modules home page,
located at http://www.zxmodules.de

-------------------------------------------------------------------------------
Cross-referenced components and libraries that are not part of this source:

- Delphi 2006 components (c) 1995-2004 Borland Software Corporation
-----------------------------------------------------------------------------}

Unit zxpformat;

Interface

Uses
  Windows, SysUtils, Classes, Graphics, ulaplus;

Type
  zxp_errors=(zxp_ok,zxp_transparentnotallowed,zxp_badimagecontent,
    zxp_missingzxpidentifier,zxp_pixelsdonotmatchattributes,
    zxp_pixelwidtherror,zxp_invalidpixelline,zxp_invalidattributeline,
    zxp_invalidtransparentarea,zxp_invalidpicturesize,zxp_invalidulapluspalette);

    PULAPlusPalette=^TULAPlusPalette;

Const
  identifier_standard='ZX-Paintbrush image';
  identifier_extended='ZX-Paintbrush extended image';

  zxp_errormessages:array[succ(low(zxp_errors))..high(zxp_errors)] Of String=
    ('Transparent ZXP images are not allowed!','Bad ZXP image content!',
     'Missing ZXP identifier!','Pixels do not match to attributes!',
     'Pixel width error!','Invalid pixel definition(s)!',
     'Invalid attribute definition(s)!','Invalid transparent area!',
     'Invalid picture size!','Invalid ULAPlus palette!');

Function readbitmapfromZXPformat(zxp:TStringList; dest:TBitmap;
  Var transparent:Boolean; ErrorMessages:Boolean; Var OutTimexMode:Boolean;
  OutPalette:PULAPlusPalette=NIL):zxp_errors;
  //Reads a ZXP file and creates a complete bitmap. If an OutPalette is given,
  //and the file contains an ULAPlus palette, it is written into the palette.

Function WriteBitmapToZXPformat(bitmap:TBitmap; Const Attrs:String; sl:TStringList;
  ErrorMessages:Boolean; Palette:TULAPlusPalette):zxp_errors;
  //Writes the pixels in the monochrome bitmap Bitmap and the attributes in the
  //string Attrs as a ZXP text file.

Implementation

Uses
  Dialogs;

Function GetLeftRightBorder(Const line:String; Var left,right:Integer;
  Var resline:String):zxp_errors;
//Detects the width of transparent pixel markers (represented by asterisks)
//on the left and right side of the pixel string.
//Results the pure pixel line (without transparent areas) in resline.
Var
  i:Integer;
Begin
  Result:=zxp_ok;
  resline:='';
  left:=0;
  right:=0;
  For i:=1 To Length(line) Do
    If line[i]='*' Then Inc(left)
      Else Break;
  For i:=Length(line) DownTo 1 Do
    If line[i]='*' Then Inc(right)
      Else Break;
  For i:=left+1 To Length(line)-right Do
    If Not(line[i] In ['1','0']) Then Begin
      Result:=zxp_invalidpixelline;
      Break;
    End; {If}
  If (Result=zxp_ok) And (Length(line)>left+right) Then
    resline:=Copy(line,left+1,Length(line)-right-left);
End;

Function GetHexBytes(Const line:String; Var resline:String):Boolean;
//results a number of hexadecimal values transformed to a character string
Var
  i,number:Integer;
  c:Char;
Begin
  Result:=True;
  resline:='';
  i:=1;
  number:=-1;
  While i<=Length(line) Do Begin
    c:=line[i];
    Case c Of
      '0'..'9':Begin
        If number<0 Then number:=0;
        number:=number*16+Ord(c)-Ord('0');
      End; {0..9}
      'a'..'f':Begin
        If number<0 Then number:=0;
        number:=number*16+Ord(c)-Ord('a')+10;
      End; {a..f}
      'A'..'F':Begin
        If number<0 Then number:=0;
        number:=number*16+Ord(c)-Ord('A')+10;
      End; {A..F}
      ' ':Begin
        If number>=0 Then Begin
          If number<=$FF Then resline:=resline+Chr(number)
            Else Begin
              Result:=False;
              Break;
            End; {else}
          number:=-1;
        End; {If}
      End; {end of number}
      Else Begin
        Result:=False;
        Break;
      End; {else}
    End; {case}
    Inc(i);
  End; {while}
  If number>=0 Then Begin
    If number<=$FF Then resline:=resline+Chr(number)
      Else Result:=False;
  End; {if}
End;

Function readoverlayfromZXPformat(zxp:TStringList; b:TBitmap; Var attrs:String;
  Var left,top,right,bottom:Integer; Var transparent:Boolean;
  Var attributeheight:Integer; Var ulapalette:String):zxp_errors;
Type
  Sections=(section_none,section_pixels,section_attributes,
    section_ulapluspalette,section_undefined);
Var
  line:String;
  toparea:Boolean;
  section:Sections;
  i,Width,Height,awidth,aheight,l,r,x,y,lastlength:Integer;
  resline:String;
  pixels:TStringList;
Begin
  Result:=zxp_ok;
  pixels:=TStringList.Create;
  Try
    section:=section_none; //No section has been reached yet.
    ulapalette:='';
    Width:=0;
    Height:=0;
    awidth:=0;
    aheight:=0;
    left:=-1;
    right:=-1;
    top:=0;
    bottom:=0;
    toparea:=True;
    attrs:='';
    lastlength:=0;
    transparent:=False;
    attributeheight:=8;
    //Convert the content of the ZXP file into the text-based string list.
    //Colour attributes are red into the Attrs string.
    For i:=0 To zxp.Count-1 Do Begin
      line:=trim(zxp[i]);
      If Length(line)=0 Then Begin
        If lastlength<>0 Then Begin
          Inc(section);
          If section=section_undefined Then Break;
        End; {if}
      End
        Else Begin
          Case section Of
            section_none:Begin
              If (line=identifier_standard) Or (line=identifier_extended) Then
                Else Begin
                  Result:=zxp_missingzxpidentifier;
                  Break
                End; {else}
            End; {section_none}
            section_pixels:Begin
              If Width=0 Then Width:=Length(line)
                Else
              If Length(line)<>Width Then Begin
                Result:=zxp_pixelwidtherror;
                Break;
              End; {if}
              Inc(Height);
              If line=stringofchar('*',Width) Then Begin
                If toparea Then Inc(top)
                  Else Inc(bottom);
              End
                Else Begin
                  toparea:=False;
                  Result:=GetLeftRightBorder(line,l,r,resline);
                  If Result<>zxp_ok Then Break;
                  If left=-1 Then left:=l
                    Else
                  If left<>l Then Begin
                    Result:=zxp_invalidtransparentarea;
                    Break;
                  End; {if}
                  If right=-1 Then right:=r
                    Else
                  If right<>r Then Begin
                    Result:=zxp_invalidtransparentarea;
                    Break;
                  End; {if}
                  pixels.Add(resline);
                End; {else}
            End; {reading pixels data}
            section_attributes:Begin
              If GetHexBytes(line,resline) Then Begin
                If awidth=0 Then awidth:=Length(resline)
                  Else
                If Length(resline)<>awidth Then Begin
                  Result:=zxp_invalidattributeline;
                  Break;
                End; {if}
                Inc(aheight);
                attrs:=attrs+resline;
              End
                Else Begin
                  Result:=zxp_invalidattributeline;
                  Break;
                End; {else}
            End; {reading attributes}
            section_ulapluspalette:Begin
              //Reading a ULAPlus palette, if present:
              If GetHexBytes(line,resline) Then ulapalette:=ulapalette+resline
                Else Begin
                  Result:=zxp_invalidulapluspalette;
                  Break;
                End; {else}
            End; {section_ulapluspalette}
          End; {case}
        End; {Else}
      lastlength:=Length(line);
    End; {for}
    If Result=zxp_ok Then Begin
      If Length(ulapalette) In [0,64] Then Begin
        If (Width=awidth*8) And ((Height=aheight) Or (Height=aheight*8)) Then Begin
          attributeheight:=Height Div aheight;
          If (Width Mod 8=0) And (Height Mod attributeheight=0) Then Begin
            transparent:=(left<>0) Or (right<>0) Or (top<>0) Or (bottom<>0);
            //creating the bitmap:
            b.setsize(Width-left-right,Height-top-bottom);
            For y:=0 To b.Height-1 Do Begin
              resline:=pixels[y];
              For x:=0 To b.Width-1 Do Begin
                If resline[x+1]='1' Then b.Canvas.pixels[x,y]:=clblack
                  Else b.Canvas.Pixels[x,y]:=clwhite;
              End; {for}
            End; {for}
          End
            Else Result:=zxp_invalidpicturesize;
        End
          Else Result:=zxp_pixelsdonotmatchattributes;
      End
        Else Result:=zxp_invalidulapluspalette;
    End; {if}
  finally
    pixels.Free;
  End; {try}
End;

Function readbitmapfromZXPformat;
Var
  source:TBitmap;
  h,x,y,scanlinewidth:Integer;
  attrs,Palette:String;
  ulapluspalette:TULAPlusPalette;
  left,right,top,bottom,inkindex,paperindex,b,attributeheight,index:Integer;
  attr,byt:Byte;
  p:^Byte;
  q:PRGBQUAD;
  a:PChar;
Begin
  source:=TBitmap.Create;
  Try
    source.PixelFormat:=pf1bit;
    Result:=readoverlayfromZXPformat(zxp,source,attrs,left,top,right,bottom,
      transparent,attributeheight,Palette);
    If Result=zxp_ok Then Begin
      OutTimexMode:=attributeheight=1;
      If Length(Palette)=64 Then SetPalette(ulapluspalette,Palette)
        Else SetStandardPalette(ulapluspalette);
      If (Not transparent) And (left=0) And (right=0) And (top=0) And (bottom=0) Then Begin
        Try
          dest.setsize(source.Width,source.Height);
          dest.PixelFormat:=pf32bit;
          h:=source.Height-1;
          scanlinewidth:=source.Width Div 8;
          For y:=0 To h Do Begin
            p:=source.ScanLine[h];
            q:=dest.ScanLine[h];
            a:=@attrs[h Div attributeheight*scanlinewidth+1];
            For x:=0 To scanlinewidth-1 Do Begin
              attr:=Ord(a^);
              GetInkPaper(attr,inkindex,paperindex);
              byt:=p^;
              Inc(p);
              For b:=7 DownTo 0 Do Begin
                If odd(byt shr b) Then index:=paperindex
                  Else index:=inkindex;
                q^:=colortoRGBQUAD(ulapluspalette[index]);
                Inc(q);
              End; {for}
              Inc(a);
            End; {for}
            Dec(h);
          End; {for}
          If Assigned(OutPalette) Then OutPalette^:=ulapluspalette;
        except
          Result:=zxp_badimagecontent;
        End; {try}
      End
        Else Result:=zxp_transparentnotallowed;
    End; {if}
  finally
    source.Free;
  End; {try}
  If ErrorMessages And (Result<>zxp_ok) Then showmessage(zxp_errormessages[Result]);
End;

Function WriteBitmapToZXPformat(bitmap:TBitmap; Const Attrs:String; sl:TStringList;
  ErrorMessages:Boolean; Palette:TULAPlusPalette):zxp_errors;
Var
  i,x,y,attributeheight:Integer;
  p:^Integer;
  q:PChar;
  standardpalette:Boolean;
  bmp:TBitmap;
  r:String;
Begin
  Result:=zxp_ok;
  bmp:=TBitmap.Create;
  Try
    bmp.Assign(bitmap);
    bmp.PixelFormat:=pf32bit;
    If Length(Attrs)=(bmp.Width Div 8)*(bmp.Height Div 8) Then attributeheight:=8
      Else
    If Length(Attrs)=(bmp.Width Div 8)*bmp.Height Then attributeheight:=1
      Else attributeheight:=0;
    If attributeheight>0 Then Begin
      standardpalette:=isstandardpalette(Palette);
      If (attributeheight=1) Or (Not standardpalette) Then sl.Add(identifier_extended)
        Else sl.Add(identifier_standard);
      sl.Add('');
      //Creating pixel data:
      For y:=0 To bmp.Height-1 Do Begin
        p:=bmp.ScanLine[y];
        SetLength(r,bmp.Width);
        For i:=1 To bmp.Width Do Begin
          If p^=0 Then r[i]:='1'
            Else r[i]:='0';
          Inc(p);
        End; {for}
        sl.Add(r);
      End; {for}
      sl.Add('');
      //Creating attributes:
      q:=@Attrs[1];
      For y:=0 To bmp.Height Div attributeheight-1 Do Begin
        r:='';
        For x:=0 To bmp.Width Div 8-1 Do Begin
          If x>0 Then r:=r+' ';
          r:=r+IntToHex(Ord(q^),2);
          Inc(q);
        End; {for}
        sl.Add(r);
      End; {for}
      If Not standardpalette Then Begin
        //Adding ULAPlus palette:
        sl.Add('');
        r:='';
        For x := 0 To 63 Do Begin
          If x>0 Then r:=r+' ';
          r:=r+IntToHex(getColor64fromRGB(Palette[x]),2);
        End; {for}
        sl.Add(r);
      End; {if}
    End
      Else Result:=zxp_pixelsdonotmatchattributes;
  finally
    bmp.Free;
  End; {try}
  If ErrorMessages And (Result<>zxp_ok) Then showmessage(zxp_errormessages[Result]);
End;

End.
